
/*
  [L U] = iluk(A)
  [L U] = iluk(A,fill_lev)

  default value:
  1               1. fill_lev = fill level
  
  INPUT PARAMETERS
  ----------------------------------
  fill_lev    : Level of fill for ILUK preconditioner
*/


#include <octave/oct.h>
#include "defun-dld.h"
#include "error.h"
#include "gripes.h"
#include "oct-obj.h"
#include "utils.h"
#include "oct-map.h"

#include "MatrixType.h"
#include "SparseCmplxLU.h"
#include "SparsedbleLU.h"
#include "ov-re-sparse.h"
#include "ov-cx-sparse.h"
#include "Sparse.h"

extern "C"{
  #include <globheads.h>
  #include <protos.h>
  #include <defs.h>
  #include <ios.h> 
}


#include <stdio.h>
#include <stdlib.h>
#include <string.h> 
#include <math.h>
#include <sys/time.h>

//ia = ridx = row index
//ja = cidx = col index

void CSC_cs(int n, int nnz, double *data, int *ridx, int *cidx, csptr bmat){
  //ridx is row indices
  int *ja = (int*)malloc(sizeof(int)*nnz);
  int *ia = (int *)malloc(sizeof(int)*nnz);
  
  int k = 0;
  for(int i=0;i<n;i++){
    for(int j=0;j<(cidx[i+1]-cidx[i]);j++){
      ja[k] = i;
      k++;
    }
  }
  
  memcpy(ia,ridx,nnz*sizeof(int));
  COOcs(n,nnz,data,ja,ia, bmat);
  
  free(ia);
  free(ja);
}


void csr_csc(int n,int nnz,double *a,int *ia,int *ja,double *ao,int *iao,int *jao){
  //csr to csc
  //csr ia = n, ja = nnz, a = nnz;
  //csc iao = nnz, jao = n, ao = nnz;

  for(int i=0;i<n+1;i++)
    jao[i] = 0;
  for(int i=0;i<nnz;i++){
    jao[ja[i]+1]++;
  }
  for(int i=0;i<n;i++){
    jao[i+1] = jao[i+1] + jao[i];
  }
  
  for(int i=0;i<n;i++){
    for(int j=ia[i];j<ia[i+1];j++){
      ao[jao[ja[j]]] = a[j];
      iao[jao[ja[j]]] = i;
      jao[ja[j]]++;
    }
  }

  for(int i=n;i>=0;i--)
    jao[i+1] = jao[i];
  jao[0] = 0;
}


void cs_CSC(double *ao, int *iao, int *jao, csptr bmat){
  //iao n+1, jao nnz
  //csc allocated array

  int nnz = nnz_cs(bmat);

  int *ia = (int*)malloc(sizeof(int)*(bmat->n+1));
  int *ja = (int*)malloc(sizeof(int)*nnz);
  double *a = (double*)malloc(sizeof(double)*nnz);

  ia[0] = 0;
  for(int i=0;i<bmat->n;i++){
    ia[i+1] = ia[i] + bmat->nzcount[i];
  }

  int k = 0;  
  for(int i=0;i<bmat->n;i++){
    for(int j=0;j<bmat->nzcount[i];j++){
      ja[k] = bmat->ja[i][j];
      a[k] = bmat->ma[i][j];
      k++;
    }
  }
  csr_csc(bmat->n,nnz,a,ia,ja,ao,iao,jao);
  free(ia);
  free(ja);
  free(a);
}


void cs_CSC_L(double *ao, int *iao, int *jao, csptr bmat){
  //iao n+1, jao nnz
  //csc allocated array

  int nnz = nnz_cs(bmat);

  int *ia = (int*)malloc(sizeof(int)*(bmat->n+1));
  int *ja = (int*)malloc(sizeof(int)*(nnz+bmat->n));
  double *a = (double*)malloc(sizeof(double)*(nnz+bmat->n));

  ia[0] = 0;
  for(int i=0;i<bmat->n;i++){
    ia[i+1] = ia[i] + bmat->nzcount[i]+1;
  }

  int k = 0;  
  for(int i=0;i<bmat->n;i++){
    for(int j=0;j<bmat->nzcount[i];j++){
      ja[k] = bmat->ja[i][j];
      a[k] = bmat->ma[i][j];
      k++;
    }
    ja[k] = i;
    a[k] = 1.0;
    k++;
  }
  csr_csc(bmat->n,nnz+bmat->n,a,ia,ja,ao,iao,jao);
  free(ia);
  free(ja);
  free(a);
}

void cs_CSC_U(double *ao, int *iao, int *jao, csptr bmat, double *diag){
  //iao n+1, jao nnz
  //csc allocated array

  int nnz = nnz_cs(bmat);

  int *ia = (int*)malloc(sizeof(int)*(bmat->n+1));
  int *ja = (int*)malloc(sizeof(int)*(nnz+bmat->n));
  double *a = (double*)malloc(sizeof(double)*(nnz+bmat->n));

  ia[0] = 0;
  for(int i=0;i<bmat->n;i++){
    ia[i+1] = ia[i] + bmat->nzcount[i]+1;
  }

  int k = 0;  
  for(int i=0;i<bmat->n;i++){
    for(int j=0;j<bmat->nzcount[i];j++){
      ja[k] = bmat->ja[i][j];
      a[k] = bmat->ma[i][j];
      k++;
    }
    ja[k] = i;
    a[k] = 1.0/diag[i];
    k++;
  }
  csr_csc(bmat->n,nnz+bmat->n,a,ia,ja,ao,iao,jao);
  free(ia);
  free(ja);
  free(a);
}

DEFUN_DLD (iluk, args, nargout,
           "  [L U] = iluk(A)\n\
  [L U] = iluk(A,fill_lev)\n\n\
  default value:\n\
  1               1. fill_lev = fill level\n\n\
  INPUT PARAMETERS\n\
  ----------------------------------\n\
  fill_lev    : Level of fill for ILUK preconditioner\n\n")
{
  octave_value_list retval;
  int nargin = args.length ();
  if (nargin == 0)
    print_usage ();
  else if (nargin > 2)
    error("iluk: incorrect number of arguments");
  else if(nargin == 1){
    if(args(0).type_name () == "sparse matrix"){
      SparseMatrix sm = args(0).sparse_matrix_value();
      int *cidx = sm.cidx();
      int *ridx = sm.ridx();
      double *data = sm.data();
      int nnz = sm.nnz();
      
      csptr csmat = (csptr)malloc(sizeof(SparMat));
      iluptr lu = (iluptr)malloc(sizeof(ILUSpar));
      CSC_cs(sm.cols(),nnz,data,ridx,cidx,csmat);
      int lfil = 1;
      /*-------------------- call ILUK preconditioner set-up  */
      int ierr = ilukC(lfil, csmat, lu, stderr);
      /*----------------------------------------------------- */
      if(ierr == -2){
        cleanILU(lu);
        error("zero diagonal element found...\n" );
      }else if(ierr != 0){
        error("*** iluk error, ierr != 0 ***\n");
        exit(-1);
      }
      SparseMatrix U(lu->U->n,lu->U->n,nnz_cs(lu->U)+lu->U->n);
      SparseMatrix L(lu->L->n,lu->L->n,nnz_cs(lu->L)+lu->L->n);
      cs_CSC_U(U.data(),U.ridx(),U.cidx(), lu->U, lu->D);
      cs_CSC_L(L.data(),L.ridx(),L.cidx(), lu->L);
      
      cleanCS(csmat);
      cleanILU(lu);
      retval(0) = octave_value(L,MatrixType(MatrixType::Lower));
      retval(1) = octave_value(U,MatrixType(MatrixType::Upper));
      return retval;
    }else{
      error("First parameter should be sparse matrix");
    }
  }else if(nargin == 2){
    if(args(0).type_name () == "sparse matrix"){
      int lfil = args(1).int_value();

      if (!error_state){
        SparseMatrix sm = args(0).sparse_matrix_value();
        int *cidx = sm.cidx();
        int *ridx = sm.ridx();
        double *data = sm.data();

        int nnz = sm.nnz();
        
        csptr csmat = (csptr)malloc(sizeof(SparMat));
        iluptr lu = (iluptr)malloc(sizeof(ILUSpar));
        CSC_cs(sm.cols(),nnz,data,ridx,cidx,csmat);
        /*-------------------- call ILUK preconditioner set-up  */
        int ierr = ilukC(lfil, csmat, lu, stderr);
        /*----------------------------------------------------- */
        if(ierr == -2){
          cleanILU(lu);
          error("zero diagonal element found...\n" );
        }else if(ierr != 0){
          error("*** iluk error, ierr != 0 ***\n");
          exit(-1);
        }
        SparseMatrix U(lu->U->n,lu->U->n,nnz_cs(lu->U)+lu->U->n);
        SparseMatrix L(lu->L->n,lu->L->n,nnz_cs(lu->L)+lu->L->n);
        cs_CSC_U(U.data(),U.ridx(),U.cidx(), lu->U, lu->D);
        cs_CSC_L(L.data(),L.ridx(),L.cidx(), lu->L);

        cleanCS(csmat);
        cleanILU(lu);
        retval(0) = octave_value(L,MatrixType(MatrixType::Lower));
        retval(1) = octave_value(U,MatrixType(MatrixType::Upper));
        return retval;
      }else{
        error("second parameter should be integer");
      }
    }else{
      error("First parameter should be sparse matrix");
    }
  }
  return retval;
}

